@@ -37,6 +37,13 @@ typedef SSIZE_T ssize_t;
 
 #include "darray.h"
 
+static inline bool
+is_aligned(const void *restrict pointer, size_t byte_count)
+{
+    return (uintptr_t) pointer % byte_count == 0;
+}
+
+
 #define ARRAY_SIZE(arr) ((sizeof(arr) / sizeof(*(arr))))
 
 #define MIN(a, b) ((a) < (b) ? (a) : (b))
@@ -106,9 +113,23 @@ istreq(const char *s1, const char *s2)
 }
 
 static inline bool
-istreq_prefix(const char *s1, const char *s2)
+istrneq(const char *s1, const char *s2, size_t len)
 {
-    return istrncmp(s1, s2, strlen(s1)) == 0;
+    return istrncmp(s1, s2, len) == 0;
+}
+
+#define istreq_prefix(s1, s2) istrneq(s1, s2, sizeof(s1) - 1)
+
+static inline char *
+strcpy_safe(char *dest, size_t size, const char *src) {
+    if (!dest || !size || !src)
+        return NULL;
+    const char * const limit = dest + size - 1;
+    while (dest < limit && *src) {
+        *dest++ = *src++;
+    }
+    *dest = '\0';
+    return *src ? NULL : dest;
 }
 
 static inline char *
