@@ -14,10 +14,12 @@
 #include <string.h>
 #include <stdbool.h>
 
-#include "keymap.h"
+#include "xkbcommon/xkbcommon.h"
 #include "darray.h"
-#include "xkbcomp-priv.h"
+#include "keymap.h"
 #include "text.h"
+#include "utils.h"
+#include "xkbcomp-priv.h"
 
 static bool
 has_unbound_vmods(struct xkb_keymap *keymap, xkb_mod_mask_t mask)
@@ -102,7 +104,7 @@ FindInterpForKey(struct xkb_keymap *keymap, const struct xkb_key *key,
     for (int s = 0; s < num_syms; s++) {
         bool found = false;
         for (darray_size_t i = 0; i < keymap->num_sym_interprets; i++) {
-            const struct xkb_sym_interpret *interp = &keymap->sym_interprets[i];
+            struct xkb_sym_interpret * const interp = &keymap->sym_interprets[i];
             xkb_mod_mask_t mods;
 
             found = false;
@@ -160,6 +162,7 @@ FindInterpForKey(struct xkb_keymap *keymap, const struct xkb_key *key,
             }
             if (found) {
                 darray_append(*interprets, interp);
+                interp->required = true;
                 break;
             }
         }
@@ -322,6 +325,22 @@ CheckMultipleActionsCategories(struct xkb_keymap *keymap, struct xkb_key *key)
     }
 }
 
+static void
+add_key_aliases(struct xkb_keymap *keymap, darray_size_t min, darray_size_t max,
+                struct xkb_key_alias *aliases)
+{
+    darray_size_t num_key_aliases = 0;
+    for (darray_size_t alias = min; alias <= max; alias++) {
+        const KeycodeMatch entry = keymap->key_names[alias];
+        if (entry.is_alias && entry.found) {
+            aliases[num_key_aliases++] = (struct xkb_key_alias) {
+                .alias = alias,
+                .real = entry.alias.real
+            };
+        }
+    }
+}
+
 /**
  * This collects a bunch of disparate functions which was done in the server
  * at various points that really should've been done within xkbcomp.  Turns out
@@ -331,6 +350,90 @@ CheckMultipleActionsCategories(struct xkb_keymap *keymap, struct xkb_key *key)
 static bool
 UpdateDerivedKeymapFields(struct xkb_keymap *keymap)
 {
+    /**
+     * Create the key alias list
+     *
+     * Since we already allocated the key name LUT, we try to overwrite it:
+     * - An alias entry is exactly 2× the size of a LUT entry (on most compilers).
+     * - We ensure that there is a real key for each alias.
+     * - So, if each alias maps to a different keycode, we usually have:
+     *   current alloc(keymap.key_names) ≥ needed alloc(keymap.key_aliases).
+     * - Usual keymaps have enough space in the LUT to enable overwriting real
+     *   key entries without overlap with the alias entries.
+     * - In the unlikely event that this is not the case, we need to realloc.
+     */
+#ifdef _MSC_VER
+    /* MSVC does not pack bit fields as efficiently as GCC and Clang */
+    static_assert(sizeof(struct xkb_key_alias) == sizeof(KeycodeMatch), "");
+#else
+    static_assert(sizeof(struct xkb_key_alias) == 2 * sizeof(KeycodeMatch), "");
+#endif
+    darray_size_t num_key_aliases = 0;
+    darray_size_t min_alias = DARRAY_SIZE_MAX;
+    darray_size_t max_alias = 0;
+    for (xkb_atom_t alias = 0; alias < keymap->num_key_names; alias++) {
+        const KeycodeMatch entry = keymap->key_names[alias];
+        if (entry.is_alias && entry.found) {
+            num_key_aliases++;
+            if (min_alias == DARRAY_SIZE_MAX)
+                min_alias = alias;
+            max_alias = alias;
+        }
+    }
+    if (num_key_aliases) {
+        /*
+         * No fancy algorithm used here: either we can trivially write the whole
+         * range of `xkb_key_aliases` without overlapping with the `KeycodeMatch`
+         * alias entries or we allocate a new array.
+         * In practice no new allocation is is needed and it delivers performs
+         * better than using a buffer to handle overlaps.
+         */
+        const darray_size_t required_space = sizeof(struct xkb_key_alias)
+                                           / sizeof(KeycodeMatch)
+                                           * num_key_aliases;
+        if (min_alias >= required_space) {
+            /* Overwrite before the *first* alias entry */
+            add_key_aliases(keymap, min_alias, max_alias, keymap->key_aliases);
+            /* Shrink */
+            struct xkb_key_alias * const r =
+                realloc(keymap->key_aliases,
+                        num_key_aliases * sizeof(*keymap->key_aliases));
+            if (r == NULL)
+                return false;
+            keymap->key_aliases = r;
+        } else if (keymap->num_key_names - max_alias - 1 > required_space) {
+            /* Overwrite after the *last* alias entry, then move to the start */
+            struct xkb_key_alias * const aliases = (struct xkb_key_alias *) (
+                keymap->key_names - required_space -
+                !is_aligned(keymap->key_names - required_space,
+                            sizeof(struct xkb_key_alias))
+            );
+            add_key_aliases(keymap, min_alias, max_alias, aliases);
+            memcpy(keymap->key_aliases, aliases,
+                   num_key_aliases * sizeof(*aliases));
+            /* Shrink */
+            struct xkb_key_alias * const r =
+                realloc(keymap->key_aliases,
+                        num_key_aliases * sizeof(*keymap->key_aliases));
+            if (r == NULL)
+                return false;
+            keymap->key_aliases = r;
+        } else {
+            /*
+             * No space trivially available: instead of processing the LUT with
+             * a buffer to handle overlaps, we simply allocate a dedicated array
+             */
+            struct xkb_key_alias * const aliases = calloc(num_key_aliases,
+                                                          sizeof(*aliases));
+            if (aliases == NULL)
+                return false;
+            add_key_aliases(keymap, min_alias, max_alias, aliases);
+            free(keymap->key_names);
+            keymap->key_aliases = aliases;
+        }
+    }
+    keymap->num_key_aliases = num_key_aliases;
+
     /* Find all the interprets for the key and bind them to actions,
      * which will also update the vmodmap. */
     struct xkb_key *key;
@@ -348,6 +451,32 @@ UpdateDerivedKeymapFields(struct xkb_keymap *keymap)
             if (key->vmodmap & (UINT32_C(1) << idx))
                 mod->mapping |= key->modmap;
 
+    if (keymap->format >= XKB_KEYMAP_FORMAT_TEXT_V2) {
+        /*
+         * XKB extension: canonical virtual modifiers (incompatible with X11)
+         *
+         * Virtual modifiers that are not mapped, map to their canonical
+         * encoding, i.e. themselves.
+         *
+         * This excludes:
+         * - Modifiers mapped *explicitly*: e.g. `virtual_modifiers M = 0x1;`.
+         * - Modifiers mapped *implicitly* via vmodmap/modifier_map, i.e. the
+         *   usual X11 way used in xkeyboard-config.
+         */
+        xkb_vmods_enumerate(idx, mod, &keymap->mods) {
+            const xkb_mod_mask_t mask = UINT32_C(1) << idx;
+            if (mod->mapping == 0 && !(keymap->mods.explicit_vmods & mask)) {
+                mod->mapping = mask;
+                /*
+                 * Make the mapping explicit, so that it is interoperable
+                 * without relying on the implementation-specific virtual
+                 * modifiers indices.
+                 */
+                keymap->mods.explicit_vmods |= mask;
+            }
+        }
+    }
+
     /* Update the canonical modifiers state mask */
     assert(keymap->canonical_state_mask == MOD_REAL_MASK_ALL);
     xkb_mod_mask_t extra_canonical_mods = 0;
@@ -434,7 +563,7 @@ CompileKeymap(XkbFile *file, struct xkb_keymap *keymap)
         if (file->file_type < FIRST_KEYMAP_FILE_TYPE ||
             file->file_type > LAST_KEYMAP_FILE_TYPE) {
             if (file->file_type == FILE_TYPE_GEOMETRY) {
-                log_vrb(ctx, 1,
+                log_vrb(ctx, XKB_LOG_VERBOSITY_BRIEF,
                         XKB_WARNING_UNSUPPORTED_GEOMETRY_SECTION,
                         "Geometry sections are not supported; ignoring\n");
             } else {
